// Chemfiles, a modern library for chemistry file reading and writing
// Copyright (C) Guillaume Fraux and contributors -- BSD license

#ifndef CHEMFILES_CAPI_H
#define CHEMFILES_CAPI_H

#include "chemfiles/capi/types.h"
#include "chemfiles/capi/misc.h"
#include "chemfiles/capi/property.h"
#include "chemfiles/capi/atom.h"
#include "chemfiles/capi/residue.h"
#include "chemfiles/capi/topology.h"
#include "chemfiles/capi/cell.h"
#include "chemfiles/capi/frame.h"
#include "chemfiles/capi/trajectory.h"
#include "chemfiles/capi/selection.h"

#endif
